;;; matrix.el	-- Implementation of matricies and matrix operations --

;; Copyright (C) 1998 Free Software Foundation, Inc.

;; Author: David Bakhash <cadet@bu.edu>
;; Maintainer: David Bakhash <cadet@bu.edu>
;; Version: 0.2
;; Created: Wed Nov 11 23:02:48 EST 1998
;; Keywords: extensions, lisp, math

;; This file is not part of emacs or XEmacs.

;; XEmacs is free software; you can redistribute it and/or modify it
;; under the terms of the GNU General Public License as published by
;; the Free Software Foundation; either version 2 of the License, or
;; (at your option) any later version.

;; XEmacs program is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
;; General Public License for more details.

;; You should have received a copy of the GNU General Public License
;; along with XEmacs; see the file COPYING.  If not, write to the Free
;; Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
;; 02111-1307, USA.

;;; Synched up with: Not in FSF.

;;; Commentary:

;; This package contains a large number of basic matrix operations.
;; Several of them were ported from James Landay's
;; <landay@cs.berkeley.edu> `matrix.lsp' (which was written in Common
;; Lisp, as part of the guesture recognition package in Garnet).
;; Since (X)Emacs Lisp does not have the power of CL-style arrays,
;; this package makes extensive use of `cl-array.el'.  The basic
;; routintes at your disposal are illustrated as follows:

;; (pprint-matrix
;;  (setq m (matrix '((1 2)
;; 		      (3 4)))))
;; 
;; 
;; 1.00000 2.00000 
;; 3.00000 4.00000 
;; 
;; (determinant m)
;; ==> -2.0
;; 
;; (ppm (matrix-inverse m)) ; `ppm' is a shorthand alias for `pprint-matrix'
;; 
;; -2.00000 1.00000 
;; 1.50000 -0.50000 
;; 
;; (matrixp m)
;; ==> t
;; 
;; (square-matrix-p m)
;; ==> t
;; 
;; (positivie-definite-matrix-p m)
;; ==> nil
;; 
;; (ppm
;;  (transpose m))
;; 
;; 1.00000 3.00000 
;; 2.00000 4.00000 
;; 
;; (ppm
;;  (setq vv (matrix-vector 2 3 4)))
;; 
;; 2.00000 
;; 3.00000 
;; 4.00000 
;; 
;; (matrix-multiply (transpose vv) vv)
;; ==> 29
;; 
;; (ppm
;;  (matrix-multiply vv (transpose vv)))
;; 
;; 4.00000 6.00000 8.00000 
;; 6.00000 9.00000 12.00000 
;; 8.00000 12.00000 16.00000 
;;
;; (ppm
;;  (eye 3))
;; 
;; 1.00000 0.00000 0.00000 
;; 0.00000 1.00000 0.00000 
;; 0.00000 0.00000 1.00000 
;; 
;; (pprint-matrix
;;  (eye 4 pi))
;; 
;; 3.14159 0.00000 0.00000 0.00000 
;; 0.00000 3.14159 0.00000 0.00000 
;; 0.00000 0.00000 3.14159 0.00000 
;; 0.00000 0.00000 0.00000 3.14159 
;; 
;; (ppm
;;  (zeros 3))
;; 
;; 0.00000 0.00000 0.00000 
;; 0.00000 0.00000 0.00000 
;; 0.00000 0.00000 0.00000 
;; 
;; 
;; (pprint-matrix
;;  (ones 2 4))
;; 
;; 1.00000 1.00000 1.00000 1.00000 
;; 1.00000 1.00000 1.00000 1.00000 


(eval-when (compile load eval)
  (require 'cl-array))

(mapc #'require '(cl cl-array))

;; we need to redefine an `=' that can take an arbitrary number of
;; arguments...
(defun matrix-cl-= (&rest args)
  (loop 
   for arg on args
   for current = (first arg)
   for next = (second arg)
   while next
   always (= current next)))

;; INVERT-MATRIX is an internal function which should not be used
;; directly.  use `matrix-inverse' and `determinant' instead !!!
;; invert-matrix inverts the in-matrix and places the result
;; in out-matrix and then returns the determinant.  The matrices
;; must be square and of the same size.
;;
;; Parameters:
;;    in-matrix  - matrix to invert
;;    out-matrix - matrix to return result in
;;
(defun* invert-matrix (in-matrix out-matrix &optional (determinant-only nil))
  (let ((dim (array-dimension in-matrix 0));; dimension of matrix
	(det 1)				; determinant of matrix
	(l nil)				; permutation vector
	(m nil)				; permutation vector
	(temp 0))       
    (if (not (= dim (array-dimension in-matrix 1)))
	(error "invert-matrix () - matrix not square"))
    (if (not (equal (array-dimensions in-matrix) 
		    (array-dimensions out-matrix)))
	(error "invert-matrix () - matrices not of the same size"))
    ;; copy in-matrix to out-matrix if they are not the same
    (when (not (equal in-matrix out-matrix))
      (do ((i 0 (1+ i)))
	  ((>= i dim))    
	(do ((j 0 (1+ j)))
	    ((>= j dim)) 
	  (setf (aref* out-matrix i j) (float (aref* in-matrix i j))))))
    ;; allocate permutation vectors for l and m, with the 
    ;; same origin as the matrix
    (setf l (make-array `(,dim)))
    (setf m (make-array `(,dim)))
    (do ((k 0 (1+ k))
	 (biga 0)
	 (recip-biga 0))
	((>= k dim))
      (setf (aref* l k) k)
      (setf (aref* m k) k)
      (setf biga (aref* out-matrix k k))
      ;; find the biggest element in the submatrix
      (do ((i k (1+ i)))
	  ((>= i dim))    
	(do ((j k (1+ j)))
	    ((>= j dim)) 
	  (when (> (abs (aref* out-matrix i j)) (abs biga))
	    (setf biga (aref* out-matrix i j))
	    (setf (aref* l k) i)
	    (setf (aref* m k) j))))
      ;; interchange rows
      (if (> (aref* l k) k)
	  (do ((j 0 (1+ j))
	       (i (aref* l k)))
	      ((>= j dim)) 
	    (setf temp (- (aref* out-matrix k j)))
	    (setf (aref* out-matrix k j) (float (aref* out-matrix i j)))
	    (setf (aref* out-matrix i j) (float temp))))
      ;; interchange columns 
      (if (> (aref* m k) k)
	  (do ((i 0 (1+ i))
	       (j (aref* m k)))
	      ((>= i dim)) 
	    (setf temp (- (aref* out-matrix i k)))
	    (setf (aref* out-matrix i k) (float (aref* out-matrix i j)))
	    (setf (aref* out-matrix i j) (float temp))))
      ;; divide column by minus pivot (value of pivot 
      ;; element is in biga)
      (if (equalp biga 0) 
	  (return-from invert-matrix 0))
      (setf recip-biga (/ 1 biga))
      (do ((i 0 (1+ i)))
	  ((>= i dim)) 
	(if (not (equal i k))
	    (setf (aref* out-matrix i k) 
		  (* (float (aref* out-matrix i k)) (- recip-biga)))))
      ;; reduce matrix
      (do ((i 0 (1+ i)))
	  ((>= i dim)) 
	(when (not (equal i k))
	  (setf temp (aref* out-matrix i k))
	  (do ((j 0 (1+ j)))
	      ((>= j dim)) 
	    (if (not (equal j k))
		(incf (aref* out-matrix i j) 
		      (* temp (aref* out-matrix k j)))))))
      ;; divide row by pivot
      (do ((j 0 (1+ j)))
	  ((>= j dim)) 
	(if (not (equal j k))
	    (setf (aref* out-matrix k j)
		  (float (* (aref* out-matrix k j) recip-biga)))))
      (setf det (* det biga))		; product of pivots
      (setf (aref* out-matrix k k) (float recip-biga))
      )					; k loop
    (unless determinant-only
      ;; final row & column interchanges
      (do ((k (1- dim) (1- k)))
	  ((< k 0))
	(if (> (aref* l k) k)
	    (do ((j 0 (1+ j))
		 (i (aref* l k)))
		((>= j dim))
	      (setf temp (aref* out-matrix j k))
	      (setf (aref* out-matrix j k) 
		    (float (- (aref* out-matrix j i))))
	      (setf (aref* out-matrix j i) (float temp))))
	(if (> (aref* m k) k)
	    (do ((i 0 (1+ i))
		 (j (aref* m k)))
		((>= i dim))
	      (setf temp (aref* out-matrix k i))
	      (setf (aref* out-matrix k i) 
		    (float (- (aref* out-matrix j i))))
	      (setf (aref* out-matrix j i) (float temp))))))
    det					; return determinant
    )
  )

(defun pprint-matrix (mat &optional stream)
  "Pretty print the matrix MAT (to the optional stream STREAM).
Note: `ppm' is an alias for this function."
  (cond ((matrixp mat)
	 (let ((m (array-dimension mat 0))
	       (n (array-dimension mat 1)))
	   (terpri)
	   (loop for i below m		; rows
	     do
	     (loop for j below n       ; columns
	       do
	       (princ (format "%7.5f " (aref* mat i j)) stream))
	     (terpri))
	   (terpri)))
	((vectorp mat)
	 (let ((m (array-dimension mat 0)))
	   (terpri)
	   (loop for i below m		; rows
	     do
	     (princ (format "%7.5f " (aref* mat i)) stream)
	     (terpri))
	   (terpri)))
	(t ; else
	 (error "pprint-matrix wants to see a matrix as it's first argument"))))

(defalias 'ppm 'pprint-matrix)

(defun matrix (rows-list)
  "Like VECTOR, but creates a matrix whose rows are listed in ROWS-LIST"
  (if (apply #'matrix-cl-= (mapcar #'length rows-list))
      (let ((m (length rows-list))
	    (n (length (car rows-list))))
	(make-array (list m n)
			  :initial-contents rows-list))
    (error "ROWS-LIST passed to function `matrix' must all be of the same length")))

(defun matrix-vector (&rest elements)
  "Like `vector', but produces a vector that the matrix package can
work with."
  (matrix (mapcar #'list elements)))

(defun* copy-matrix (mat &key dimensions-only)
  "Return the copy of matrix MAT, and all its elements, unless
the `DIMENSIONS-ONLY' key is non-nil so that the copied matrix
shares only the dimensions of MAT, but dimensions are all nil"
  (if (matrixp mat)
      (let ((mat-copy (make-array (array-dimensions mat))))
	(unless dimensions-only
	  ;; copy all the elements from MAT
	  (dotimes (i (array-total-size mat)) 
	    (setf (row-major-aref mat-copy i)
		  (row-major-aref mat i))))
	mat-copy)))

(defun* matrix-multiply (mat1 &rest mats &aux (mat2 (car mats)))
  "Compute the matrix product: (MAT1)*(MAT2)*...*(MATn).
Of course, the no. of columns in MAT1 must = no. of rows of MAT2"
  ;; This should, one day, except regular rank one vectors too...
  (cond ((and (vector-or-matrix-p mat1)
	      (null mat2))
	 ;; only one matrix, so return the identity
	 ;; (copy-matrix mat1))
	 (if (= 1 (array-total-size mat1))
	     ;; matrix is really a scalar
	     (row-major-aref mat1 0)
	   mat1))
	((and (matrixp mat1)
	      (matrixp mat2))
	 (let ((m1 (array-dimension mat1 0)) ; rows of MAT1
	       (n1 (array-dimension mat1 1)) ; columns of MAT1
	       (m2 (array-dimension mat2 0)) ; rows of MAT2
	       (n2 (array-dimension mat2 1))) ; columns of MAT2
	   (if (= n1 m2)		; check dimensions
	       (let ((product (make-array (list m1 n2))))
		 (loop for i below m1
		   do
		   (loop for k below n2
		     do
		     (setf (aref* product i k)
			   (loop for j below n1
			     sum (* (aref* mat1 i j)
				    (aref* mat2 j k))))))
		 (apply #'matrix-multiply product (cdr mats)))
	     (error "matricies don't have the right dimensions for multiplication"))))
	((vectorp mat2)
	 (apply #'matrix-multiply
		mat1
		(make-array (list (length mat2) 1)
						   :initial-contents (map 'list #'list mat2))
		(cdr mats)))
	(t (error "Wrong type arguments to function `matrix-multiply'"))))

(defun* matrix-add (mat1 &rest mats &aux (mat2 (car mats)))
  "Compute the matrix sum: (MAT1)+(MAT2)+...+(MATn).
Of course, the no. of columns in MAT1 must = no. of rows of MAT2"
  (cond ((and (matrixp mat1)
	      (null mat2))
	 mat1)
	((and (matrixp mat1)
	      (matrixp mat2))
	 (if (and (equal (array-dimensions mat1)
			 (array-dimensions mat2)))
	     (let ((sum (make-array (array-dimensions mat1))))
	       (loop for i below (array-total-size mat1)
		 do
		 (setf (row-major-aref sum i)
		       (+ (row-major-aref mat1 i)
			  (row-major-aref mat2 i))))
	       (apply #'matrix-add sum (cdr mats)))
	   (error "matricies don't have the right dimensions for addition")))
	(t (error "Wrong type arguments to function `matrix-add'"))))

(defun vector-or-matrix-p (mat)
  (and (arrayp* mat)
       (<= (array-rank mat) 2)))

(defun matrixp (mat)
  (and (arrayp* mat)
       (= (array-rank mat) 2)))

(defun square-matrix-p (mat)
  (and (matrixp mat)
       (apply #'matrix-cl-= (array-dimensions mat))))

(defun symmetric-matrix-p (mat)
  (and (square-matrix-p mat)
       (let ((dim (array-dimension mat 0)))
	 (loop for i from 1 below dim
	     always
	       (loop for j below i
		   always
		     (= (aref* mat i j)
			(aref* mat j i)))))))       

(defun positivie-definite-matrix-p (mat &optional semi-definite-p)
  "Non-nil iff matrix MAT is positive definite.
if SEMI-DEFINITE-P is specified and non-nil, then return t iff MAT >= 0."
  (let ((inequality (if semi-definite-p
			#'>=
		      #'>)))
    (and (symmetric-matrix-p mat)
	 (plusp (aref* mat 0 0))
	 (let ((dim (array-dimension mat 0)))
	   (loop for i from 1 below dim
	     always (funcall inequality
			     (determinant (submatrix mat 
						     (list 0 i)
						     (list 0 i)))
			     0))))))

(defun transpose (mat)
  "Given a matrix MAT, return the transpose"
  (cond ((matrixp mat)
	 (let* ((m (array-dimension mat 0)) ; rows
		(n (array-dimension mat 1)) ; columns
		(tp (make-array (list n m))))
	   (loop for i below m
	     do
	     (loop for j below n
	       do
	       (setf (aref* tp j i)
		     (aref* mat i j))))
	   tp))
	((vectorp mat)
	 (let* ((length (array-dimension mat 0))
		(tp (make-array (list 1 length))))
	   (loop for i below length
	     do
	     (setf (row-major-aref tp i) (row-major-aref mat i)))
	   tp))
	(t (error "transpose needs to see a vector or matrix"))))

(defun matrix-inverse (mat)
  ;; #### RENAME `inverse' ??? ####
  "Return the inverse of matrix MAT"
  (if (square-matrix-p mat)
      (let ((inverse (make-array (array-dimensions mat))))
	(if (zerop (invert-matrix mat inverse))
	    ;; singular matrix!
	    (error "Tried to invert a singular matrix")
	  inverse))
    (error "argument to MATRIX-INVERSE must be a square matrix")))

;;; #### This can definately be optimized !!! ####
(defun determinant (mat)
  "Return the determinant of matrix MAT.
Note that Lisp vectors are interpreted as column vectors in this
package, and they cannot be transposed with this function."
  (if (square-matrix-p mat)
      (let ((inverse (make-array (array-dimensions mat))))
	;; note that `invert-matrix' actually returns the determinant
	(invert-matrix mat inverse t))
    (error "argument to MATRIX-INVERSE must be a square matrix")))

(defun* eye (size &optional scale-1 &key scale)
  "Make an identity matrix size x size.
Optional SCALE (default=1) is the diagonal element.
Keyword argument :SCALE is also acceptable."
  (let ((mat (make-array (list size size) :initial-element 0)))
    (setq scale (or scale scale-1 1))
    (do ((i 0 (1+ i)))
	((>= i size))
      (setf (aref* mat i i) scale))
    mat))

(defun* zeros (size &optional (cols size))
  "Make an zero matrix of size SIZE
If COLS is also specified, then you get a SIZE by COLS matrix of
zeros"
  (make-array (list size cols) :initial-element 0))

(defun* ones (size &optional (cols size) &key (scale 1))
  "Make a matrix all of whos entries are 1.
Optional argument SCALE will set all entries to SCALE"
  (make-array (list size cols) :initial-element scale))

;; now we move on to extracting parts of vectors/matricies...

(defun submatrix (mat &rest limits)
  "Create a submatrix of matrix MAT.
The matrix will have limits specified by LIMITS These can either be
lists: (min max|:all) or atoms.  In the latter case, the atom can be
number or the symbol `:all', which means to include all (or rest) of
elements.
This function is settable."
  (setq limits (mapcar #'(lambda (l)
			   (cond ((atom l)
				  (cond ((integerp l)
					 (list l l))
					((eq l :all)
					 (list 0 :all))
					(t ; else
					 (error "illegal limit"))))
				 (t ; (listp l)
				  l)))
		       limits))
  (loop
    with dims = (array-dimensions mat)
    for l in-ref limits
    and i upfrom 0
    do
    (cond ((eq l :all)
	   (setf l (list 0 (1- (nth i dims)))))
	  (t ; (listp l)
	   (when (eq (second l) :all)
	     (setf (second l) (1- (nth i dims)))))))		     
  (destructuring-bind ((a b) (c d)) limits
    (loop 
      with submat = (make-array (list (1+ (- b a))
				      (1+ (- d c))))
      for i from a to b
      and p upfrom 0
      do
      (loop 
	for j from c to d
	and q upfrom 0
	do
	(setf (aref* submat p q) (aref* mat i j)))
      finally
      (return submat))))

(defun matrix-set-submatrix (mat limits submat)
  (setq limits (mapcar #'(lambda (l)
			       (cond ((atom l)
				      (cond ((integerp l)
					     (list l l))
					    ((eq l :all)
					     (list 0 :all))
					    (t ; else
					       (error "illegal limit"))))
				     (t ; (listp l)
				      l)))
			   limits))
  (loop
    with dims = (array-dimensions mat)
    for l in-ref limits
    and i upfrom 0
    do
    (cond ((eq l :all)
	   (setf l (list 0 (1- (nth i dims)))))
	  (t ; (listp l)
	   (when (eq (second l) :all)
	     (setf (second l) (1- (nth i dims)))))))		     
  (destructuring-bind ((a b) (c d)) limits
    (loop
      for i from a to b
      and p upfrom 0
      do
      (loop 
	for j from c to d
	and q upfrom 0
	do
	(setf (aref* mat i j) (aref* submat p q))))
    submat))

(defsetf submatrix (mat &rest limits) (submat)
  `(matrix-set-submatrix ,mat (list ,@limits) ,submat))

(provide 'matrix)
